@@ -60,6 +60,7 @@ def interact_with_build_environment(osbase, instance_dir, machine_name, *,
                        flags=nspawn_flags,
                        tmp_apt_cache_dir=aptcache_tmp,
                        pkginjector=pkginjector,
+                       syscall_filter=osbase.global_config.syscall_filter,
                        verbose=True)
 
     if source_pkg_dir:
@@ -177,9 +178,6 @@ def internal_execute_build(osbase, pkg_dir, build_only=None, *,
             prep_flags = ['--build-prepare']
             if build_only == 'arch':
                 prep_flags.append('--arch-only')
-            if qa_lintian:
-                # install Lintian from the start, if we run it later
-                prep_flags.append('--lintian')
             r = nspawn_run_helper_persist(osbase,
                                           instance_dir,
                                           machine_name,
@@ -197,6 +195,7 @@ def internal_execute_build(osbase, pkg_dir, build_only=None, *,
                             '-u', 'builder',
                             '--private-network']
             helper_flags = ['--build-run']
+            helper_flags.extend(['--suite', osbase.suite])
             if buildflags:
                 helper_flags.append('--buildflags={}'.format(';'.join(buildflags)))
             r = nspawn_run_helper_persist(osbase,
@@ -206,7 +205,8 @@ def internal_execute_build(osbase, pkg_dir, build_only=None, *,
                                           '/srv',
                                           nspawn_flags=nspawn_flags,
                                           tmp_apt_cache_dir=aptcache_tmp,
-                                          pkginjector=pkginjector)
+                                          pkginjector=pkginjector,
+                                          syscall_filter=osbase.global_config.syscall_filter)
             # exit, unless we are in interactive mode
             if r != 0 and not interact:
                 return False
