@@ -23,6 +23,37 @@ from .utils import temp_dir, print_error, print_warn, print_info, safe_run, run_
 from .utils.env import colored_output_allowed, unicode_allowed
 
 
+__systemd_version = None
+
+
+def systemd_version():
+    global __systemd_version
+    if __systemd_version:
+        return __systemd_version
+
+    __systemd_version = -1
+    try:
+        out, _, _ = safe_run(['systemd-nspawn', '--version'])
+        parts = out.split(' ', 2)
+        if len(parts) >= 2:
+            __systemd_version = int(parts[1])
+    except Exception as e:
+        print_warn('Unable to determine systemd version: {}'.format(e))
+
+    return __systemd_version
+
+
+def systemd_version_atleast(expected_version: int):
+    v = systemd_version()
+    # we always assume we are running the highest version,
+    # if we failed to determine the right systemd version
+    if v < 0:
+        return True
+    if v >= expected_version:
+        return True
+    return False
+
+
 def get_nspawn_personality(osbase):
     '''
     Return the syszemd-nspawn container personality for the given combination
@@ -36,7 +67,7 @@ def get_nspawn_personality(osbase):
     return None
 
 
-def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
+def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[], syscall_filter=[]):
     '''
     Execute systemd-nspawn with the given parameters.
     Mess around with cgroups if necessary.
@@ -46,16 +77,28 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
 
     capabilities = []
     full_dev_access = False
+    full_proc_access = False
+    ro_kmods_access = False
+    all_privileges = False
     for perm in allow_permissions:
         perm = perm.lower()
-        if perm.startswith('cap_'):
-            capabilities.append(perm.upper())
-        elif perm == 'full-dev-access':
+        if perm.startswith('cap_') or perm == 'all':
+            if perm == 'all':
+                capabilities.append(perm)
+                print_warn('Container retains all privileges.')
+                all_privileges = True
+            else:
+                capabilities.append(perm.upper())
+        elif perm == 'full-dev':
             full_dev_access = True
+        elif perm == 'full-proc':
+            full_proc_access = True
+        elif perm == 'read-kmods':
+            ro_kmods_access = True
         else:
             print_info('Unknown allowed permission: {}'.format(perm))
 
-    if (capabilities or full_dev_access) and not osbase.global_config.allow_unsafe_perms:
+    if (capabilities or full_dev_access or full_proc_access) and not osbase.global_config.allow_unsafe_perms:
         print_error('Configuration does not permit usage of additional and potentially dangerous additional permissions. Exiting.')
         sys.exit(9)
 
@@ -63,8 +106,18 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
     cmd.extend(['-M', machine_name])
     if full_dev_access:
         cmd.extend(['--bind', '/dev'])
+        if systemd_version_atleast(244):
+            cmd.append('--console=pipe')
+    if full_proc_access:
+        cmd.extend(['--bind', '/proc'])
+        if not all_privileges:
+            print_warn('Container has access to host /proc')
+    if ro_kmods_access:
+        cmd.extend(['--bind-ro', '/lib/modules/'])
     if capabilities:
         cmd.extend(['--capability', ','.join(capabilities)])
+    if syscall_filter:
+        cmd.extend(['--system-call-filter', ' '.join(syscall_filter)])
     cmd.extend(parameters)
 
     if not full_dev_access:
@@ -79,7 +132,8 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
             # child process - edit the cgroup to allow full access to all
             # devices. Hopefully there won't be too much need for this awful code.
             parent_pid = os.getppid()
-            print_warn('Giving container access to all host devices.')
+            if not all_privileges:
+                print_warn('Container will have direct access to all host devices.')
 
             syscg_devices_allow = '/sys/fs/cgroup/devices/machine.slice/machine-{}.scope/devices.allow'.format(escaped_full_machine_name)
             tries = 0
@@ -110,7 +164,7 @@ def _execute_sdnspawn(osbase, parameters, machine_name, allow_permissions=[]):
 
 
 def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=[], *,
-                       tmp_apt_cache_dir=None, pkginjector=None, allowed=[], verbose=False):
+                       tmp_apt_cache_dir=None, pkginjector=None, allowed=[], syscall_filter=[], verbose=False):
     if isinstance(command, str):
         command = command.split(' ')
     if isinstance(flags, str):
@@ -135,7 +189,7 @@ def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=
         osbase.aptcache.create_instance_cache(aptcache_tmp_dir)
 
         # run command in container
-        ret = _execute_sdnspawn(osbase, params, machine_name, allowed)
+        ret = _execute_sdnspawn(osbase, params, machine_name, allowed, syscall_filter)
 
         # archive APT cache, so future runs of this command are faster
         osbase.aptcache.merge_from_dir(aptcache_tmp_dir)
@@ -151,7 +205,7 @@ def nspawn_run_persist(osbase, base_dir, machine_name, chdir, command=[], flags=
     return ret
 
 
-def nspawn_run_ephemeral(osbase, base_dir, machine_name, chdir, command=[], flags=[], allowed=[]):
+def nspawn_run_ephemeral(osbase, base_dir, machine_name, chdir, command=[], flags=[], allowed=[], syscall_filter=[]):
     if isinstance(command, str):
         command = command.split(' ')
     if isinstance(flags, str):
@@ -167,7 +221,7 @@ def nspawn_run_ephemeral(osbase, base_dir, machine_name, chdir, command=[], flag
     params.extend(['-aqxD', base_dir])
     params.extend(command)
 
-    return _execute_sdnspawn(osbase, params, machine_name, allowed)
+    return _execute_sdnspawn(osbase, params, machine_name, allowed, syscall_filter)
 
 
 def nspawn_make_helper_cmd(flags):
@@ -195,7 +249,7 @@ def nspawn_run_helper_ephemeral(osbase, base_dir, machine_name, helper_flags, ch
 
 
 def nspawn_run_helper_persist(osbase, base_dir, machine_name, helper_flags, chdir='/tmp', *,
-                              nspawn_flags=[], tmp_apt_cache_dir=None, pkginjector=None, allowed=[]):
+                              nspawn_flags=[], tmp_apt_cache_dir=None, pkginjector=None, allowed=[], syscall_filter=[]):
     cmd = nspawn_make_helper_cmd(helper_flags)
     return nspawn_run_persist(osbase,
                               base_dir,
@@ -205,4 +259,5 @@ def nspawn_run_helper_persist(osbase, base_dir, machine_name, helper_flags, chdi
                               nspawn_flags,
                               tmp_apt_cache_dir=tmp_apt_cache_dir,
                               pkginjector=pkginjector,
-                              allowed=allowed)
+                              allowed=allowed,
+                              syscall_filter=syscall_filter)
